#include "AirportSim.h"
#include <OpenGL/gl.h>
#include <iostream>
#include "3DMath.h"
#include "Random.h"
#include "Timing.h"
#include <queue>
#include <list>
#include <cmath>

using namespace std;

Airport airport;

Random r;

///////////////////
//	DISTRIBUTIONS
///////////////////

double LinearDist::getNumber() {
	return r.rangedFloat(low,high);
}

double ArrivalDist::getNumber() {
	double num_p = r.rangedFloat(0.0, 0.999999999999);
	double num = num_p * 40;
	return( arrival_dist_lookup[int(num)]*(1.0-num_p) + arrival_dist_lookup[int(num + 1.0)]*num_p );
}

double ArrivalDist::max() {
	return 170;
}

double ArrivalDist::arrival_dist_lookup[41] = {
		0, 21.25, 28.05, 32.98, 37.23, 40.8, 44.03, 47.09, 49.81, 52.53, 55.08, 57.46,
		59.84, 62.22, 64.43, 66.64, 68.85, 71.06, 73.1, 75.31, 77.35, 79.56, 81.6, 83.81,
		85.85, 88.06, 90.27, 92.65, 94.86, 97.24, 99.79, 102.34, 105.06, 107.78, 110.84, 
		114.24, 117.81, 122.06, 127.33, 134.47, 170 };


/* Jay's addition */
//if peak is not < 0, power must be odd
WaitDist::WaitDist(float min, float max, float peak, int power)
		:min_(min), max_(max), peak_(peak), power_(power), range_(max_ - min_) 

{
	ret_range_h_ = ( (max_ - peak_) * pow( range_ / (max_ - peak_) , power_) );
	ret_range_l_ = ( (min_ - peak_) * pow( range_ / (min_ - peak_) , power_) );
}

double WaitDist::getNumber()
{
	float num = r.rangedFloat( (min_ - peak_) / range_ , (max_ - peak_) / range_ );
	if (num < 0)
		num = pow(num, power_)*ret_range_l_;
	else num = pow(num, power_)*ret_range_h_;
	return( num + peak_ );
}

double SpeedDist::getNumber() {
	double num_p = r.rangedFloat(0.0, 0.999999999999);
	double num = num_p * 50;
	return( speed_dist_lookup[int(num)]*(1.0-num_p) + speed_dist_lookup[int(num + 1.0)]*num_p );
}

double SpeedDist::speed_dist_lookup[51] = {
	0.304, 0.7296912,0.778764, 0.8257032,0.874776, 0.8939784,0.9025128,  0.9110472,0.9195816, 0.9259824,0.9345168,
	0.9430512,0.9515856, 0.96012,0.9686544, 0.9750552,0.9835896, 0.992124,1.0006584, 1.0091928,1.0177272,
	1.0262616,1.0326624, 1.0411968,1.0497312, 1.0582656,1.0689336, 1.0945368,1.12014, 1.1457432,1.1713464,
	1.1969496,1.2225528, 1.2524232,1.2972288, 1.344168,1.3889736, 1.4273784,1.4529816,1.4764512, 1.4999208,1.5233904,
	1.5489936,1.5724632, 1.5959328,1.6450056, 1.6983456,1.749552, 1.888236,2.0290536, 2.1314664 };

/* end Jay's addition */



ArrivalDist arrivalDist;

WaitDist checkinWait(2, 20, 5, 3);
WaitDist securityWait(2, 20, 3, 5);
WaitDist bathroomWait(1, 15, 4, 3);

SpeedDist baseSpeedDistribution;
SpeedDist disabledBaseSpeedDistribution;
SpeedDist escortBaseSpeedDistribution;