/*
 *  AirportSim.cpp
 *  AirportSim
 *
 *  Created by Troy Lawlor on 2/2/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include "AirportSim.h"
#include <OpenGL/gl.h>
#include <iostream>
#include "3DMath.h"
#include "Random.h"
#include "Timing.h"
#include <queue>
#include <list>
#include <fstream>
#include <typeinfo> // needed for typeid

using namespace std;

string itos( int num ) {
	char buffer [33];
	sprintf(buffer,"%d",num);

	return string( buffer );
}

Random r;

Airport airport;

extern float percentNeedingEscort;
extern float percentNeedingMotorized;

float minCheckinTime = 30;
float bodyWidth = 1;

extern int numDollys;

int numCheckinLines = 30;
int numSecurityLines = 5;
int numBathroomLines = 5;

int initialRegularChairsPerBank = 200;
int initialMotorChairsPerBank = 200;

LinearDist arrivalFlightFull(0.3,1.0);
LinearDist arrivalFlightConnection(2,5);
LinearDist flightCapacity(30,120);
float arrivalFlightNeedConnectionPercent = 25;
float percentDisabled = 5.5;

float escortSalary = 13;
extern int numDollys;

int totalEscorts=15;

int biasSet=-3;

float gTime=-3;

int gDays=0;
int gIteration=0;

//////// STAT COLLECTING
float timeSinceLastMinute=0;
std::vector<int> totalWheelchairsAvailable;

int disabledMissedFlight=0;
int totalDisabledPassengersSpawned=0;
int totalPassengersSpawned=0;

ofstream statFile("stats.csv");

///////// COSTS

float wheelchairCost = 0.10;
float motorWheelchairCost = 0.17;
float motorBatteryCost = 0.23;
float complaintsCost = 60;
float missedFlightCost = 256;
float dollyCost = 0.035;

float totalCost=0;
	
///////// ETC

extern float timeBias(float t);
extern float capacityBias(float c);

void Airport::tweakParameters() {
	/*totalEscorts=r.rangedInt(1,35);
	initialRegularChairsPerBank=r.rangedInt(10,300);
	initialMotorChairsPerBank=r.rangedInt(10,300);
	numDollys=r.rangedInt(2,16);*/
	//initialRegularChairsPerBank += biasSet*4;
	//initialMotorChairsPerBank += biasSet*4;
}

void Airport::readParameters() {
	ifstream in("parameters.txt");
	std::string trash;
	in >> trash >> numCheckinLines;
in >> trash >> numSecurityLines;
in >> trash >> numBathroomLines;
in >> trash >> minCheckinTime;

in >> trash >> percentDisabled;
in >> trash >> percentNeedingEscort;
in >> trash >> percentNeedingMotorized;

in >> trash >> totalEscorts;
in >> trash >> escortSalary;

in >> trash >> initialRegularChairsPerBank;
in >> trash >> initialMotorChairsPerBank;
in >> trash >> numDollys;

in >> trash >> wheelchairCost;
in >> trash >> motorWheelchairCost;
in >> trash >> motorBatteryCost;
in >> trash >> complaintsCost;
in >> trash >> missedFlightCost;
in >> trash >> dollyCost;

in >> trash >> bodyWidth;


}

void Airport::writeParameters() {
	string name = itos(gIteration)+".txt";
	ofstream out(name.c_str());
	out << "Escorts: " << totalEscorts << endl;
	out << "Wheelchairs: " << initialRegularChairsPerBank << endl;
	out << "Initial Wheelchairs: " << initialMotorChairsPerBank << endl;
	out << "Num Dollys: " << numDollys << endl;
}

///////////////
//  GRAPH
///////////////


void Graph::addEdge(int a, int b, int weight) {
	verts[a].adj.push_back(b);
	verts[a].weight.push_back(weight);
}

void Graph::addBiEdge(int a, int b) {
	verts[a].adj.push_back(b);
	verts[a].weight.push_back(FREE);
	verts[b].adj.push_back(a);
	verts[b].weight.push_back(FREE);
}

void Graph::addEdges(int a, int low, int high) {
	for(int i=low; i<=high; ++i) {
		addBiEdge(a,i);
	}
}

void Graph::addVertex() {
	verts.push_back(Vertex());
}

std::vector<int>& Graph::adjacent(int i) {
	return verts[i].adj;
}

std::vector<int>& Graph::weight(int i) {
	return verts[i].weight;
}

int Graph::totalVertices() {
	return verts.size();
}


///////////////
//  AREA
///////////////


Area::Area(int p_type, float p_size, float p_x, float p_y) : type(p_type), numPeople(0), distance(p_size), x(p_x), y(p_y) {
	const float TO_METER = 0.3048f;
	const float SEPARATION_BIAS = 3.5f;
	distance *= TO_METER;
	x*=TO_METER;
	y*=TO_METER;
	x*=SEPARATION_BIAS;
	y*=SEPARATION_BIAS;
	maxPeople = (int)((distance*distance)/(bodyWidth*bodyWidth));
	if(type==CHECKIN) { waitProcess.lines.resize(numCheckinLines); waitProcess.numDist = checkinWait; }
	else if(type==SECURITY_IN) { waitProcess.lines.resize(numSecurityLines); waitProcess.numDist = securityWait; }
	else if(type==BATHROOM) { waitProcess.lines.resize(numBathroomLines); waitProcess.numDist = bathroomWait; }
}

bool Area::enter() {
	if(congestion()<1.0f) {
		numPeople++;
		return true;
	}
	else return false;
}

void Area::leave() {
	numPeople--;
}

float Area::congestion() {
	return (float)numPeople/(float)maxPeople;  //  should be some other model, exponential, etc.
}


// NEED NUMBERS
float Area::speedModifier() {
	return 1.0f-(congestion()/1.0f)+0.1f;
}

void Area::draw() {
	float brightness = 1.0f-congestion();
	
	float border=1.0f;
	
	if(type!=FREE) {
	switch(type) {
		case START: glColor3f(0.3f,0.3f,0.3f); break;
		case BAGGAGE: glColor3f(0.7f,0.3f,0.5f); break;
		case BATHROOM: glColor3f(0.0f,1.0f,0.0f); break;
		case WC: glColor3f(1.0f,0.0f,1.0f); break;
		case CHECKIN: glColor3f(1.0f,1.0f,0.0f); break;
		case SECURITY_IN: glColor3f(1.0f,0.0f,0.0f); break;
		default: glColor3f(1.0f,0.5f,0.2f); break;
	}
	glVertex2f((x-(distance/2+border))*windowScale,(y-(distance/2+border))*windowScale);
	glVertex2f((x+(distance/2+border))*windowScale,(y-(distance/2+border))*windowScale);
	glVertex2f((x+(distance/2+border))*windowScale,(y+(distance/2+border))*windowScale);
	glVertex2f((x-(distance/2+border))*windowScale,(y+(distance/2+border))*windowScale);
	}
	
	glColor3f(1.0f,brightness,brightness);  //  color gets more red as more congested
	glVertex2f((x-distance/2)*windowScale,(y-distance/2)*windowScale);
	glVertex2f((x+distance/2)*windowScale,(y-distance/2)*windowScale);
	glVertex2f((x+distance/2)*windowScale,(y+distance/2)*windowScale);
	glVertex2f((x-distance/2)*windowScale,(y+distance/2)*windowScale);
}

void Area::think(double tm) {
	waitProcess.think(tm);
}

bool Area::mustWait(int myID, WaitProcess * &wp) {
	if(type==CHECKIN || type==SECURITY_IN) { waitProcess.add(myID); wp = &waitProcess; return true; }
	else return false;
}


///////////////
//  WHEELCHAIR BANK
///////////////


Wheelchair::Wheelchair(bool p_motorized) : motorized(p_motorized), distance(0), timeInUse(0) {
	baseSpeed = 0.5;
}

void Wheelchair::resetStats() { distance=0; timeInUse=0; }

WheelchairBank::WheelchairBank(int p_area) {
	area = p_area;
	for(int i=0; i<initialRegularChairsPerBank; ++i) {
		WC.push_back(Wheelchair(false));
	}
	for(int i=0; i<initialMotorChairsPerBank; ++i) {
		motorWC.push_back(Wheelchair(true));
	}
}

Wheelchair WheelchairBank::takeOne(bool motorized) {
	if(!motorized) {
		Wheelchair wc = WC.front();
		WC.pop_front();
		return wc;
	}
	else {
		Wheelchair wc = motorWC.front();
		motorWC.pop_front();
		return wc;
	}
}

void WheelchairBank::returnOne(Wheelchair wc) {
	if(wc.motorized)
		motorWC.push_back(wc);
	else
		WC.push_back(wc);
}

bool WheelchairBank::has(bool motorized) {
	if(motorized)
		return motorWC.empty()==false;
	else
		return WC.empty()==false;
}

int WheelchairBank::total() {
	return WC.size() + motorWC.size();
}

void WheelchairBank::resetStats() {
	int i;
	for(i=0; i<WC.size(); ++i) WC[i].resetStats();
	for(i=0; i<motorWC.size(); ++i) motorWC[i].resetStats();
}

///////////////
//  WAIT Process
///////////////


void WaitProcess::think(double tm) {
	for(int i=0; i<lines.size(); ++i) {
		if(!lines[i].waiters.empty()) {
			lines[i].timeLeft -= tm;
			if(lines[i].timeLeft<=0) {
				doneWaiting.push_back(lines[i].waiters.front());
				lines[i].waiters.pop_front();
				lines[i].timeLeft = numDist.getNumber();
			}
		}
	}
}

bool WaitProcess::isDone(int gid) {
	vector<int>::iterator i=find(doneWaiting.begin(), doneWaiting.end(), gid);
	if(i!=doneWaiting.end()) { doneWaiting.erase(i); return true; }
	else return false;
}

void WaitProcess::add(int gid) {
	//  find line with fewest people
	int min = lines[0].waiters.size();
	int which = 0;
	for(int i=1; i<lines.size(); ++i) {
		if(lines[i].waiters.size()<min) {
			min = lines[i].waiters.size();
			which = i;
		}
	}
	lines[which].waiters.push_back(gid);
}

void WaitProcess::remove(int gid) {
	vector<int>::iterator i=find(doneWaiting.begin(), doneWaiting.end(), gid);
	if(i!=doneWaiting.end()) { doneWaiting.erase(i); }
	for(int j=0; j<lines.size(); ++j) {
		deque<int>::iterator i=find(lines[j].waiters.begin(), lines[j].waiters.end(), gid);
		if(i!=lines[j].waiters.end()) { lines[j].waiters.erase(i); return;}
	}
}

///////////////
//  AIRPORT
///////////////


Airport::Airport() {
}

void Airport::init() {
	statFile.close();
	string name = itos(gIteration)+".csv";
	statFile.open(name.c_str());
	gTime=-3;
	gDays=0;
	timeSinceLastMinute=0;
	gPassengerID=0;
	gEscortID=0;
	lastWriteout=0;
	gFlightID=0;
	int i;
	graph.addVertex();  areas.push_back(Area(Area::START,  150,  0,0));//0
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100,  -50,25));//1
	graph.addVertex();  areas.push_back(Area(Area::FREE,  150,  50,25));//2
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100,  -100,75));//3
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100,  0,75));//4
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100,  100,75));//5
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100,  200,75));//6
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100,  300,75));//7
	graph.addVertex();  areas.push_back(Area(Area::CHECKIN,  200,  0,210));//8
	
	//  Concourse A/B
	graph.addVertex();  areas.push_back(Area(Area::SECURITY_IN,  150,  -100,310));//9
	graph.addVertex();  areas.push_back(Area(Area::FREE,  275,  200,510));//10
	graph.addVertex();  areas.push_back(Area(Area::FREE,  125,  100,610));//11
	graph.addVertex();  areas.push_back(Area(Area::FREE,  275,  0,510));//12
	graph.addVertex();  areas.push_back(Area(Area::FREE,  275,  -300,510));//13
	graph.addVertex();  areas.push_back(Area(Area::BATHROOM,  50,  -400,510));//14
	graph.addVertex();  areas.push_back(Area(Area::BATHROOM,  50,  100,510));//15
	
	//  Gates A
	for(i=0; i<8; ++i) {
	graph.addVertex();  areas.push_back(Area(Area::GATE1+i,  275,  200,510));//16-23
	}
	
	graph.addVertex();  areas.push_back(Area(Area::GATE1+9-1,  275,  0,510));//24
	graph.addVertex();  areas.push_back(Area(Area::GATE1+10-1,  275,  0,510));//25
	
	//  Gates B
	for(i=11-1; i<=11-1+5; ++i) {
	graph.addVertex();  areas.push_back(Area(Area::GATE1+i,  275,  -300,510));//26-31
	}
	
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100, -250 ,75));//32
	graph.addVertex();  areas.push_back(Area(Area::BAGGAGE,  140, -250 ,150));//33
	graph.addVertex();  areas.push_back(Area(Area::BAGGAGE,  75, 300 ,150));//34
	graph.addVertex();  areas.push_back(Area(Area::FREE,  100, 400 ,75));//35
	graph.addVertex();  areas.push_back(Area(Area::SECURITY_IN,  75, 500 ,75));//36
	
	//  Concourse C
	graph.addVertex();  areas.push_back(Area(Area::FREE,  125, 500 ,150));//37
	graph.addVertex();  areas.push_back(Area(Area::FREE,  200, 700 ,75));//38
	graph.addVertex();  areas.push_back(Area(Area::BATHROOM,  50, 650 ,170));//39
	graph.addVertex();  areas.push_back(Area(Area::BATHROOM,  50, 720 ,200));//40
	
	for(i=16-1; i<=21-1; ++i) {
		graph.addVertex();  areas.push_back(Area(Area::GATE1+i,  125,  500,150));//41-46
	}
	
	for(i=22-1; i<=24-1; ++i) {
		graph.addVertex();  areas.push_back(Area(Area::GATE1+i,  200,  700,75));//47-49
	}
	
	//  Wheelchair banks
	graph.addVertex();  areas.push_back(Area(Area::WC,  50, 200,125));//50
	//graph.addVertex();  areas.push_back(Area(Area::WC,  50, -200,310));//51
	wheelchairBanks.push_back(WheelchairBank(50));
	//wheelchairBanks.push_back(WheelchairBank(51));
	
	graph.addBiEdge(0,1);
	graph.addBiEdge(0,2);
	graph.addBiEdge(1,3);
	graph.addBiEdge(3,4);
	graph.addBiEdge(4,5);
	graph.addBiEdge(5,6);
	graph.addBiEdge(6,7);
	graph.addBiEdge(2,7);
	graph.addBiEdge(4,8);
	
	//  Concourse A/B
	graph.addEdge(3,9,Graph::SECURITY_IN);
	graph.addEdge(9,3,Graph::SECURITY_OUT);
	graph.addBiEdge(9,13);
	graph.addBiEdge(13,14);
	graph.addBiEdge(9,10);
	graph.addBiEdge(10,11);
	graph.addBiEdge(11,12);
	graph.addBiEdge(10,15);
	
	//  Gates A/B
	graph.addEdges(10,16,23);
	graph.addEdges(12,24,25);
	graph.addEdges(13,26,31);
	
	//  more
	graph.addBiEdge(3,32);
	graph.addBiEdge(32,33);
	graph.addBiEdge(7,34);
	graph.addBiEdge(7,35);
	graph.addEdge(35,36,Graph::SECURITY_IN);
	graph.addEdge(36,35,Graph::SECURITY_OUT);
	
	//  Concourse C
	graph.addBiEdge(36,37);
	graph.addBiEdge(36,38);
	graph.addBiEdge(37,39);
	graph.addBiEdge(38,40);
	graph.addEdges(37,41,46);
	graph.addEdges(38,47,49);
	
	//  WC
	graph.addBiEdge(6,50);
	//graph.addBiEdge(9,51);
	
	//  setup departures
	for(i=0; i<23; ++i) {
		departures.push_back(Flight(timeTillNextFlight.getNumber()+arrivalDist.max()+r.rangedFloat(0,240), gFlightID++, i, 30));
	}
	
	for(i=0; i<23; ++i) {
		arrivals.push_back(Flight(timeTillNextFlight.getNumber()+r.rangedFloat(0,240), i+256, i, 30));
	}
	
	for(i=0; i<totalEscorts; ++i)
		escorts.push_back(Escort(gEscortID++, 0, escortSalary));
}

float Airport::die() {
	gateAreas.clear();
	departures.clear();
	arrivals.clear();
	graph = Graph();
	areas.clear();
	escorts.clear();
	for(int i=0; i<passengers.size(); ++i)
		delete passengers[i];
	passengers.clear();
	wheelchairBanks.clear();
	floatingWheelchairs.clear();
	gIteration++;
	float ptc = totalCost;
	totalCost=0;
	biasSet++;
	return ptc;
}

void Airport::step(double tm) {
	int i;
	gTime += tm/60/60;
	
	if(gTime>=24) { gTime-=(int)gTime; gDays++; }
	
	//  think about flight departures
	for(i=0; i<departures.size(); ++i) {
		departures[i].minutes-=tm/60.0f;
		if(departures[i].minutes<=0) {
			removePassengers(gateArea(departures[i].gate), departures[i].number);
			int gate = departures[i].gate;
			departures.erase(departures.begin()+i);
			i--;
			departures.push_back(Flight(timeBias(timeTillNextFlight.getNumber())+arrivalDist.max(), gFlightID++, gate, capacityBias(flightCapacity.getNumber())));
		}
		//  spawn passengers
		else {
			if(departures[i].minutes<=arrivalDist.max() && departures[i].capacity>0) {
				spawnPassengers(entrance(), departures[i], departures[i].capacity);
				departures[i].capacity=0;
			}
		}
	}
	//  think about flight arrivals
	for(i=0; i<arrivals.size(); ++i) {
		arrivals[i].minutes-=tm/60.0f;
		if(arrivals[i].minutes<=0) {
			spawnPassengersArrival(gateArea(arrivals[i].gate), arrivals[i].gate, capacityBias(arrivals[i].capacity*arrivalFlightFull.getNumber()));
			int gate = arrivals[i].gate;
			arrivals.erase(arrivals.begin()+i);
			i--;
			departures.push_back(Flight(timeBias(timeTillNextFlight.getNumber()+arrivalDist.max()), (gFlightID++)+256, gate, capacityBias(flightCapacity.getNumber())));
		}
	}
	
	for(i=0; i<passengers.size(); ++i) {
		if(passengers[i]->state==Passenger::LEAVING && passengers[i]->curArea==entrance()) {
			passengers[i]->die();
			delete passengers[i];
			passengers.erase(passengers.begin()+i); i--;
		}
		else 
			passengers[i]->think(tm);
	}
	for(i=0; i<escorts.size(); ++i) {
		escorts[i].think(tm);
	}
	for(i=0; i<areas.size(); ++i) {
		areas[i].think(tm);
	}
	
	//// STAT COLLECTING
	timeSinceLastMinute+=tm;
	if(timeSinceLastMinute>60) {
		int total=0;
		timeSinceLastMinute-=60;
		for(i=0; i<wheelchairBanks.size(); ++i) {
			total+=wheelchairBanks[i].total();
		}
		totalWheelchairsAvailable.push_back(total);
	}
	
	lastWriteout+=tm;
	if(lastWriteout>60*60) {
		lastWriteout-=60*60;
		outputStatistics(statFile,60*60);
	}
}

void Airport::draw() {
	int i;
	
	//  draw edges
	glBegin(GL_LINES);
		for(int u=0; u<graph.totalVertices(); ++u) {
			for(int v=0; v<graph.verts[u].adj.size(); ++v) {
				switch(graph.verts[u].weight[v]) {
					case Graph::FREE: glColor3f(0.8f,0.8f,1.0f);	break;
					case Graph::SECURITY_IN: glColor3f(1.0f,0.0f,0.0f);	break;
					case Graph::SECURITY_OUT: glColor3f(0.0f,0.0f,1.0f);	break;
				}
				glVertex2f(areas[u].x*windowScale,areas[u].y*windowScale);
				glVertex2f(areas[graph.verts[u].adj[v]].x*windowScale,areas[graph.verts[u].adj[v]].y*windowScale);
			}
		}
	glEnd();

	glBegin(GL_QUADS);
	for(i=0; i<areas.size(); ++i) {
		areas[i].draw();
	}
	glEnd();
	
	//  draw passengers
	for(i=0; i<passengers.size(); ++i) {
		passengers[i]->draw();
	}
	
	for(i=0; i<escorts.size(); ++i) {
		escorts[i].draw();
	}
}

void Airport::debugOutput() {
	int i,j;
	/*for(int i=0; i<passengers.size(); ++i) {
		if(typeid(*passengers[i])==typeid(DisabledPassenger)) {
		cout << i << ": ";
		passengers[i]->debugOutput();
		cout << endl;
		}
	}*/
	for(i=0; i<escorts.size(); ++i) {
		cout << i << ": ";
		escorts[i].debugOutput();
		cout << endl;
	}
	
	cout << "Wheelchairs available: " << wheelchairBanks[0].WC.size() << endl;
	cout << "Motor Wheelchairs available: " << wheelchairBanks[0].motorWC.size() << endl;
	cout << "Floating wheelchairs: " << floatingWheelchairs.size() << endl;
	int totalWheelchairs=0;
	int escortWheelchairs=0;
	int passWheelchairs=0;
	for(i=0; i<passengers.size(); ++i) {
		DisabledPassenger * dp= dynamic_cast < DisabledPassenger* > (airport.passengers[i]);
		if(dp!=NULL)
			if(!dp->wheelchair.empty()) { totalWheelchairs++; passWheelchairs++;}
	}
	for(i=0; i<escorts.size(); ++i) {
		for(j=0; j<escorts[i].wheelchairs.size(); ++j) {
			totalWheelchairs++;
			escortWheelchairs++;
		}
	}
	for(i=0; i<wheelchairBanks.size(); ++i) {
		for(j=0; j<wheelchairBanks[i].WC.size(); ++j) {
			totalWheelchairs++;
		}
		for(j=0; j<wheelchairBanks[i].motorWC.size(); ++j) {
			totalWheelchairs++;
		}
	}
	cout << "Escort wheelchairs: " << escortWheelchairs << endl;
	cout << "Passenger wheelchairs: " << passWheelchairs << endl;
	cout << "Total wheelchairs: " << totalWheelchairs << endl;
	/*printFlights();*/
}

int Airport::gateArea(int which) {
	for(int i=0; i<areas.size(); ++i) {
		if(areas[i].type==Area::GATE1+which) return i;
	}
	cout << "ERROR: Gate area error!" << endl;
	return 0;
}

int Airport::entrance() {
	return 0;
}

void Airport::spawnPassengers(int area, Flight &flight, int number) {
	for(int i=0; i<number; ++i) {
		if(r.rangedFloat(0,100)<percentDisabled) {
			passengers.push_back(new DisabledPassenger(gPassengerID++, flight, area));
			totalDisabledPassengersSpawned++;
		}
		else
			passengers.push_back(new Passenger(gPassengerID++, flight, area));
			
		passengers.back()->timeToSpawn = (arrivalDist.max()-arrivalDist.getNumber())*60;
		passengers.back()->timeSpawned = passengers.back()->timeToSpawn;
		
		totalPassengersSpawned++;
	}
}

void Airport::spawnPassengersArrival(int area, int gate, int number) {
	for(int i=0; i<number; ++i) {
		Flight flight;
		//  generate a connection?
		if(r.rangedFloat(0,100)<arrivalFlightNeedConnectionPercent) {
			flight = departures[arrivalFlightConnection.getNumber()];
		}
		else {
			flight.gate = -1;
		}
		if(r.rangedFloat(0,100)<percentDisabled) {
			passengers.push_back(new DisabledPassenger(gPassengerID++, flight, area));
			totalDisabledPassengersSpawned++;
		}
		else {
			passengers.push_back(new Passenger(gPassengerID++, flight, area));
		}
		passengers.back()->state=Passenger::ARRIVED;
		totalPassengersSpawned++;
	}
}


void Airport::removePassengers(int area, int flightNumber) {
	int total=0;
	for(int i=0; i<passengers.size(); ++i) {
		if(passengers[i]->curArea==area && passengers[i]->flightID==flightNumber) { passengers[i]->die(); delete passengers[i]; passengers.erase(passengers.begin()+i); i--; total++; }
		//  make all other passengers leave
		else if(passengers[i]->flightID==flightNumber && passengers[i]->state!=Passenger::LEAVING && passengers[i]->state>Passenger::TO_CHECK_IN) {
			if(typeid(*passengers[i])==typeid(DisabledPassenger) && passengers[i]->timeSpawned/60>=minCheckinTime && passengers[i]->curArea!=-1 && passengers[i]->state!=Passenger::SPAWNED) {
				//cout << "Missed, spawned: " << passengers[i]->timeSpawned/60 << endl;
				disabledMissedFlight++;
			}
			passengers[i]->leave();
		}
	}
}

void Airport::printFlights() {
	cout << "------- DEPARTURES --------" << endl;
	for(int i=0; i<departures.size(); ++i) {
		cout << "FLIGHT: " << departures[i].number << " TIME LEFT: " << departures[i].minutes << " GATE: " << departures[i].gate << endl;
	}
}

Passenger* Airport::passenger(int pid) {
	for(int i=0; i<passengers.size(); ++i)
		if(passengers[i]->myID==pid) return passengers[i];

	cout << "ERROR: Invalid passenger with id: " << pid << endl;
	return NULL;
}

Flight Airport::departure(int which) {
	for(int i=0; i<departures.size(); ++i)
		if(departures[i].number==which) return departures[i];
	cout << "ERROR: Invalid departure with id: " << which << endl;
	return Flight();
}

WheelchairBank& Airport::wcBank(int area) {
	for(int i=0; i<wheelchairBanks.size(); ++i)
		if(wheelchairBanks[i].area==area) return wheelchairBanks[i];
	cout << "ERROR: Invalid wheelchair bank requested!" << endl;
	return wheelchairBanks[0];
}

void Airport::outputStatistics( std::ofstream &out, float totalTime ) {
	int i,j;
	
	//  wheelchair busy time
	float complaints = 0;
	int totalWheelchairs=0;
	int totalMotorWheelchairs=0;
	int totalDisabled=0;
	float wheelchairTime=0;
	float wheelchairDist=0;
	float motorWheelchairTime=0;
	float motorWheelchairDist=0;
	float waitTime=0;
	float escortedTime=0;
	float collectorTime=0;
	float escortTime = 0;
	for(i=0; i<passengers.size(); ++i) {
		DisabledPassenger * dp= dynamic_cast < DisabledPassenger* > (airport.passengers[i]);
		if(dp!=NULL) {
			totalDisabled++;
			if(!dp->wheelchair.empty()) {
				if(dp->wheelchair[0].motorized) {
					totalMotorWheelchairs++;  motorWheelchairTime += dp->wheelchair[0].timeInUse;  motorWheelchairDist+=dp->wheelchair[0].distance;
				} else {
					totalWheelchairs++;  wheelchairTime += dp->wheelchair[0].timeInUse;  wheelchairDist+=dp->wheelchair[0].distance;
				}
				dp->wheelchair[0].resetStats();
			}
			waitTime += dp->totalEscortWait;
			complaints += complaint(dp->totalEscortWait);
			escortedTime += dp->timeEscorted;
		}
		//passengers[i]->resetStats();
	}
	for(i=0; i<escorts.size(); ++i) {
		for(j=0; j<escorts[i].wheelchairs.size(); ++j) {
			if(!escorts[i].wheelchairs[j].motorized) {
				totalWheelchairs++;
				wheelchairTime += escorts[i].wheelchairs[j].timeInUse;
				wheelchairDist+= escorts[i].wheelchairs[j].distance;
			}
			else {
				totalMotorWheelchairs++;
				motorWheelchairTime += escorts[i].wheelchairs[j].timeInUse;
				motorWheelchairDist+= escorts[i].wheelchairs[j].distance;
			}
			escorts[i].wheelchairs[j].resetStats();
		}
		collectorTime += escorts[i].collectorTime;
		escortTime += escorts[i].busyTime;
		escorts[i].resetStats();
	}
	for(i=0; i<wheelchairBanks.size(); ++i) {
		for(j=0; j<wheelchairBanks[i].WC.size(); ++j) {
			totalWheelchairs++;
			wheelchairTime += wheelchairBanks[i].WC[j].timeInUse;
			wheelchairDist += wheelchairBanks[i].WC[j].distance;
		}
		for(j=0; j<wheelchairBanks[i].motorWC.size(); ++j) {
			totalMotorWheelchairs++;
			motorWheelchairTime += wheelchairBanks[i].motorWC[j].timeInUse;
			motorWheelchairDist += wheelchairBanks[i].motorWC[j].distance;
		}
		wheelchairBanks[i].resetStats();
	}
	
	float meanWCBusy = wheelchairTime/(float)totalWheelchairs/totalTime;
	float meanMWCBusy = motorWheelchairTime/(float)totalMotorWheelchairs/totalTime;
	
	//  sum costs
	float cost=0;
	cost += totalWheelchairs*wheelchairCost;
	cost += totalMotorWheelchairs*motorWheelchairCost + totalMotorWheelchairs*meanMWCBusy*motorBatteryCost;
	cost += complaints*complaintsCost;
	cost += disabledMissedFlight*missedFlightCost;
	cost += escortSalary*escorts.size();
	cost += numDollys*dollyCost;
	
	out << cost << ",";
	out << totalWheelchairs*wheelchairCost << ",";
	out << totalMotorWheelchairs*motorWheelchairCost + totalMotorWheelchairs*meanMWCBusy*motorBatteryCost << ",";
	out << complaints*complaintsCost << ",";
	out << disabledMissedFlight*missedFlightCost << ",";
	out << escortSalary*escorts.size() << ",";
	out << numDollys*dollyCost << ",";
	
	out << meanWCBusy << ",";	//  mean WC percentage of time busy
	out << meanMWCBusy << ",";	//  mean WC percentage of time busy
	out << escortTime/(float)escorts.size()/totalTime << ",";	//  mean escort percentage of time busy
	out << escortedTime/(float)totalDisabled << ",";	//  mean time disabled spent with escort
	out << collectorTime/(float)escorts.size()/totalTime << ",";  //  mean percentage of time collecting
	out << disabledMissedFlight << ",";	//  num missed flights
	out << totalPassengersSpawned << ",";
	out << totalDisabledPassengersSpawned << endl;
	
	totalCost += cost;
	
	disabledMissedFlight=0;
	totalPassengersSpawned=0;
	totalDisabledPassengersSpawned=0;
}