#include "AirportSim.h"
#include <OpenGL/gl.h>
#include <iostream>
#include "3DMath.h"
#include "Random.h"
#include "Timing.h"
#include <queue>
#include <list>

using namespace std;


extern Random r;
extern Airport airport;

float percentNeedingEscort=25;
float percentNeedingMotorized=50;

int numDollys=5;

///////////////
//  PERSON
///////////////

Person::Person() : distance(0) {}

class Node {
	public:
		int parent;
		int cost;
	
	Node() : parent(-1), cost(MAX_INT) {}
};

//  Dijkstra's to find shortest path
int Person::computeVisitList(bool byType, int start, int destinationType) {
	int i;
	Graph &graph = airport.graph;
	std::vector<Area> &areas = airport.areas;
	std::vector<Node> nodes(graph.totalVertices());
	nodes[start].cost = 0;
	vector<int> verts;
	for(i=0; i<graph.totalVertices(); ++i)
		verts.push_back(i);
		
	vector<int> visitedVerts;
	
	while(!verts.empty()) {
		//  find closest vert
		int min = MAX_INT;
		int u, index;
		for(i=0; i<verts.size(); ++i) {
			if(nodes[verts[i]].cost<min) { u = verts[i];  min = nodes[verts[i]].cost; index=i; }
		}
		if(min==MAX_INT) { cout << "Dijkstra error" << endl; return 0; }
		verts.erase(verts.begin()+index);
		
		if( (byType && areas[u].type==destinationType) || (!byType && u==destinationType) ) {
			int dest = u;
			toVisit.clear();
			while(u!=-1) {
				if(u!=start)
					toVisit.push_back(u);
				u = nodes[u].parent;
			}
			return dest;
		}
		
		visitedVerts.push_back(u);
		
		for(i=0; i<graph.adjacent(u).size(); ++i) {
			int v = graph.adjacent(u)[i];
			if(nodes[v].cost > nodes[u].cost + getEdgeWeight(graph.weight(u)[i])) {
				nodes[v].cost = nodes[u].cost + getEdgeWeight(graph.weight(u)[i]);
				nodes[v].parent = u;
			}
		}
	}
	
	return 0;
}

int Person::getEdgeWeight(int type) {
	return 1;
}

void Person::movement(double tm) {
	distance += baseSpeed*airport.areas[curArea].speedModifier()*tm;
	while(!toVisit.empty() && distance>=airport.areas[curArea].distance) {
		distance -= airport.areas[curArea].distance;
		if(airport.areas[toVisit.back()].enter()) {
			airport.areas[curArea].leave();
			curArea = toVisit.back();
			toVisit.pop_back();
			extraEnter();
		}
	}
}

void Person::extraEnter() {
}


///////////////
//  PASSENGER
///////////////


//  2 miles/hour = 0.89 meters/sec
Passenger::Passenger(int p_myID, Flight &flight, int p_startArea) : myID(p_myID), flightID(flight.number), destGate(flight.gate), state(SPAWNED), waiting(false), waitProcess(NULL), limbo(false) {
	curArea = -1;
	startArea = p_startArea;
	baseSpeed = baseSpeedDistribution.getNumber();
	destination = 0;
	if(flight.minutes<minCheckinTime) leave();
	x = r.rangedFloat(-1,1);
	y = r.rangedFloat(-1,1);
			waitTime = 0;
			timeToSpawn = 0;
		timeEscorted=0;
}

int Passenger::getEdgeWeight(int type) {
	if(type==Graph::FREE) return 1;
	else if(type==Graph::SECURITY_IN && state!=TO_GATE) return 100;
	else if(type==Graph::SECURITY_OUT && state!=TO_BAGGAGE) return 100;
	return 1;
}

void Passenger::handleState(double tm) {
	if(curArea==destination) {
		int destinationType=-1;
		switch(state) {
			case SPAWNED:
				destinationType = Area::CHECKIN;
				state=TO_CHECK_IN;
			break;
			case TO_CHECK_IN:
				destinationType = Area::GATE1+destGate;
				state=TO_GATE;
			break;
			case TO_GATE:
				if(curArea==destination) {
					/*if(airport.departure(flightID).minutes>15) {
						state=TO_BATHROOM;
						destinationType = Area::BATHROOM;
					}*/
				}
			break;
			case TO_BATHROOM:
				if(curArea==destination) {
					state=TO_GATE;
					destinationType = Area::GATE1+destGate;
				}
			break;
			case ARRIVED:
				if(destGate!=-1) { state=TO_GATE; destinationType = Area::GATE1+destGate; }
				else { state=TO_BAGGAGE; destinationType = Area::BAGGAGE; }
			break;
			case TO_BAGGAGE:
				if(curArea==destination) {
					leave();
				}
			break;
			case LEAVING:
				if(curArea==destination) {
					destinationType = Area::START;
				}
			break;
		};
		if(destinationType>-1) destination = computeVisitList(true, curArea, destinationType);
	}
}

void Passenger::think(double tm) {
	if(limbo) { timeEscorted+=tm; return; }
	
	Graph &graph = airport.graph;
	std::vector<Area> &areas = airport.areas;
	
	//  waiting to spawn in
	if(curArea==-1) {
		timeToSpawn -= tm;
		if(timeToSpawn<0) {
			if(areas[startArea].enter()) {
				curArea = startArea;
				destination = curArea;
			}
			else return;
		}
		else return;
	}
	
	if(waiting) {
		waitTime += tm;
		if(waitProcess->isDone(myID)) { waiting=false;  waitProcess=NULL; }
	}
	else {
		handleState(tm);
		if(!toVisit.empty())
			movement(tm);
	}
}

void Passenger::draw() {
	if(curArea==-1) return;
	float centerx = airport.areas[curArea].x+airport.areas[curArea].distance/2*x;
	float centery = airport.areas[curArea].y+airport.areas[curArea].distance/2*y;
	glColor3f(0.0f,0.0f,1.0f);
	const float scale=0.3;
	glBegin(GL_LINES);
		glVertex2f((centerx-scale)*windowScale, (centery-scale)*windowScale);
		glVertex2f((centerx+scale)*windowScale, (centery+scale)*windowScale);
		glVertex2f((centerx+scale)*windowScale, (centery-scale)*windowScale);
		glVertex2f((centerx-scale)*windowScale, (centery+scale)*windowScale);
	glEnd();
}

void Passenger::debugOutput() {
	if(state!=0) {
	cout << " cur: " << curArea << " dest: " << destination << " state: " << state << " speed: " << baseSpeed << " distance: " << distance;
	}
}

void Passenger::die() {
	airport.areas[curArea].leave();
}

void Passenger::leave() {
	state = LEAVING;
	destination=curArea;
	if(waiting) {
		waiting = false;
		waitProcess->remove(myID);
	}
}

void Passenger::extraEnter() {
	if((airport.areas[curArea].type==Area::CHECKIN && state==TO_CHECK_IN))
		if(airport.areas[curArea].mustWait(myID,waitProcess)) waiting=true;
}

///////////////
//  DISABLED PASSENGER
///////////////


DisabledPassenger::DisabledPassenger(int p_myID, Flight &flight, int p_startArea) : Passenger(p_myID, flight, p_startArea) {
	totalEscortWait = 0;
	baseSpeed = disabledBaseSpeedDistribution.getNumber();
	needsEscort = r.rangedFloat(0,100)<percentNeedingEscort;
	wantsMotorized = r.rangedFloat(0,100)<percentNeedingMotorized;
}

void DisabledPassenger::handleState(double tm) {
	if(curArea==destination) {
		int destinationType=-1;
		switch(state) {
			case SPAWNED:
				destinationType = Area::CHECKIN;
				state=TO_CHECK_IN;
			break;
			case TO_CHECK_IN:
				//destinationType = Area::SECURITY_IN;
				state=NEED_ESCORT;
			break;
			case TO_GATE:
			break;
			case LEAVING:
				if(curArea==destination) {
					destinationType = Area::START;
				}
			break;
			case NEED_ESCORT:
				
			break;
			case ARRIVED:
				state = NEED_ESCORT;
			break;
			case TO_BAGGAGE:
				if(curArea==destination) {
					dumpWheelchair();
					leave();
				}
			break;
			case WAITING_FOR_ESCORT:
			break;
			case INDEPENDENT:
				if(destGate!=-1) {
				destinationType = Area::GATE1+destGate;
				state=TO_GATE;
				}
				else {
				destinationType = Area::BAGGAGE;
				state=TO_BAGGAGE;
				}
			break;
		};
		if(destinationType>-1) destination = computeVisitList(true, curArea, destinationType);
	}
	switch(state) {
		case NEED_ESCORT:
		case WAITING_FOR_ESCORT:
			totalEscortWait+=tm;
		break;
	}
	
	if(!waiting && !toVisit.empty() && !wheelchair.empty())
		wheelchair[0].distance += baseSpeed*airport.areas[curArea].speedModifier()*tm;
	if(!wheelchair.empty()) wheelchair[0].timeInUse+=tm;
}

void DisabledPassenger::die() {
	Passenger::die();
	dumpWheelchair();
}

void DisabledPassenger::dumpWheelchair() {
	if(!wheelchair.empty()) {
		wheelchair.front().area = curArea;
		airport.floatingWheelchairs.push_back(wheelchair.front());
		wheelchair.clear();
	}
}

int DisabledPassenger::pickedUpByEscort(int p_escortID) {	//  returns type of area they want to go
	escortID = p_escortID;
	airport.areas[curArea].leave();
	curArea=-1;
	limbo = true;
	if(destGate!=-1)
		return Area::GATE1+destGate;
	else
		return Area::BAGGAGE;
}

void DisabledPassenger::droppedOffByEscort(int area) {
	curArea=-1;
	startArea=area;
	if(destGate!=-1)
	state = TO_GATE;
	else {
	destination=curArea;
	state = TO_BAGGAGE;
	}
	limbo = false;
}

void DisabledPassenger::draw() {
	if(curArea==-1) return;
	float centerx = airport.areas[curArea].x+airport.areas[curArea].distance/2*x;
	float centery = airport.areas[curArea].y+airport.areas[curArea].distance/2*y;
	glColor3f(1.0f,0.0f,0.0f);
	glBegin(GL_LINES);
		glVertex2f((centerx-1)*windowScale, (centery-1)*windowScale);
		glVertex2f((centerx+1)*windowScale, (centery+1)*windowScale);
		glVertex2f((centerx+1)*windowScale, (centery-1)*windowScale);
		glVertex2f((centerx-1)*windowScale, (centery+1)*windowScale);
	glEnd();
}

void DisabledPassenger::debugOutput() {
	if(state!=0) {
	//cout << "Disabled, has waited: " << totalEscortWait/60 << " minutes.";
	if(!wheelchair.empty())
	cout << "wc ";
	else
	cout << "nc ";
	Passenger::debugOutput();
	}
}

void DisabledPassenger::receiveWheelchair(Wheelchair wc) {
	wheelchair.push_back(wc);
}

void DisabledPassenger::resetStats() {
	for(int i=0; i<wheelchair.size(); ++i) wheelchair[i].resetStats();
		timeEscorted = 0;
		totalEscortWait = 0;
}


///////////////
//  ESCORT
///////////////


Escort::Escort(int p_myID, int p_startArea, float p_wages) : Person() {
	curArea=-1;  startArea = p_startArea; 
	timeSinceLastPayment = 0;
	wages = p_wages;
	cash += wages;
	myID = p_myID;
	state = SPAWNED;
	hasDisabled=false;
	x = r.rangedFloat(-1,1);
	y = r.rangedFloat(-1,1);
	baseSpeed = escortBaseSpeedDistribution.getNumber();
	collectorTime=0;
	busyTime=0;
	lastState=0;
}

void Escort::think(double tm) {
	int i;
	Graph &graph = airport.graph;
	std::vector<Area> &areas = airport.areas;
	
	for(i=0; i<wheelchairs.size(); ++i) {
		wheelchairs[i].timeInUse+=tm;
	}
	
	//  waiting to spawn in
	if(curArea==-1) {
		if(areas[startArea].enter()) {
			curArea = startArea;
			destination = curArea;
			
		}
		else return;
	}
	
	timeSinceLastPayment+=tm;
	if(timeSinceLastPayment>60*60)
		cash += wages;
	
	int destinationType=-1;
	switch(state) {
		case SPAWNED:
			destinationType = Area::WC;
			state = WAIT_FOR_CALL;
		break;
		case WAIT_FOR_CALL:
			if(lastState==GATHER_WCS) collectorTime+=tm;
			if(curArea==destination) {
				lastState = WAIT_FOR_CALL;
				WheelchairBank & bank = airport.wcBank(curArea);
				if(hasDisabled) { cout << "ERROR: Escort stole disabled person!!!!!" << endl; }
				
				//  if we have wheelchairs, deposit them
				for(i=0; i<wheelchairs.size(); ++i) {
					bank.returnOne(wheelchairs.back());
					wheelchairs.pop_back();
				}
				bool foundSomeone=false;
				//  scan for unhelped disabled people
				for(i=0; i<airport.passengers.size(); ++i) {
					if(airport.passengers[i]->state==DisabledPassenger::NEED_ESCORT) {
						DisabledPassenger * dp= dynamic_cast < DisabledPassenger* > (airport.passengers[i]);
						if(dp==NULL) { cout << "ERROR: Dynamic cast failed in TO_DISABLED!" << endl; return; }
						bool motorized = dp->wantsMotorized;
						if(bank.has(motorized)) {
							wheelchairs.push_back(bank.takeOne(motorized));
							dp->state=DisabledPassenger::WAITING_FOR_ESCORT;
							state = TO_DISABLED;
							disabledID = dp->myID;
							destination = computeVisitList(false, curArea, dp->curArea);
							foundSomeone=true;
							i=airport.passengers.size();
						}
						//  sorry no wheelchairs!
					}
				}
				//  go find unused wheelchairs
				if(!foundSomeone && !airport.floatingWheelchairs.empty()) {
					//  is somebody else on gather duty?
					int otherGatherers=0;
					for(i=0; i<airport.escorts.size(); ++i)
						if(airport.escorts[i].state==GATHER_WCS) otherGatherers++;
					if(otherGatherers<numDollys) {
						destination = computeVisitList(false, curArea, airport.floatingWheelchairs.front().area);
						state = GATHER_WCS;
					}
				}
			}
		break;
		case TO_DISABLED:
			if(curArea==destination) {
				DisabledPassenger * dp= dynamic_cast < DisabledPassenger* > (airport.passenger(disabledID));
				if(dp==NULL) { state = WAIT_FOR_CALL;  destinationType=Area::WC;  return; }
				
				if(dp->needsEscort) {
					destinationType = dp->pickedUpByEscort(myID);
					hasDisabled=true;
					state = TO_DESTINATION;
				}
				else {
					dp->receiveWheelchair(wheelchairs.back());
					dp->state = DisabledPassenger::INDEPENDENT;
					wheelchairs.pop_back();
					destinationType = Area::WC;
					state = WAIT_FOR_CALL;
				}
			}
		break;
		case TO_DESTINATION:
			if(curArea==destination) {
				DisabledPassenger * dp= dynamic_cast < DisabledPassenger* > (airport.passenger(disabledID));
				if(dp==NULL) { cout << "ERROR: Dynamic cast failed in TO_DESTINATION!" << endl; return; }
				dp->droppedOffByEscort(curArea);
				hasDisabled=false;
				state = RETURN_WC;
				destinationType = Area::WC;
			}
		break;
		case RETURN_WC:
			if(curArea==destination) {
				state = WAIT_FOR_CALL;
			}
		break;
		case GATHER_WCS:
			lastState=GATHER_WCS;
			collectorTime+=tm;
			if(airport.floatingWheelchairs.empty()) {
				destinationType = Area::WC;
				state = WAIT_FOR_CALL;
			}
			else {
				if(destination==airport.floatingWheelchairs.front().area) {
					if(destination==curArea) {
					//  gather all floaters in this area
					for(i=0; i<airport.floatingWheelchairs.size(); ++i) {
						if(airport.floatingWheelchairs[i].area==curArea) {
							wheelchairs.push_back(airport.floatingWheelchairs[i]);
							airport.floatingWheelchairs.erase(airport.floatingWheelchairs.begin()+i);
							i--;
						}
					}
					}
				}
				else if(!airport.floatingWheelchairs.empty()) {
					destination = computeVisitList(false, curArea, airport.floatingWheelchairs.front().area);
				}
				//  all done, return to home base
				else {
					destinationType = Area::WC;
					state = WAIT_FOR_CALL;
				}
			}
		break;
	}
	if(destinationType>-1) destination = computeVisitList(true, curArea, destinationType);
	
	if(!(state==WAIT_FOR_CALL && curArea==destination)) busyTime += tm;
	
	
	if(!toVisit.empty()) {
		for(int i=0; i<wheelchairs.size(); ++i)
			wheelchairs[i].distance += baseSpeed*airport.areas[curArea].speedModifier()*tm;
		movement(tm);
	}
}

void Escort::draw() {
	if(curArea==-1) return;
	float centerx = airport.areas[curArea].x+airport.areas[curArea].distance/2*x;
	float centery = airport.areas[curArea].y+airport.areas[curArea].distance/2*y;
	glColor3f(1.0f,0.0f,1.0f);
	glBegin(GL_LINES);
		glVertex2f((centerx-2)*windowScale, (centery-2)*windowScale);
		glVertex2f((centerx+2)*windowScale, (centery+2)*windowScale);
		glVertex2f((centerx+2)*windowScale, (centery-2)*windowScale);
		glVertex2f((centerx-2)*windowScale, (centery+2)*windowScale);
	glEnd();
}


void Escort::debugOutput() {
	cout << "State: " << state << " Area: " << curArea << " Destination: " << destination << " Wheelchairs: " << wheelchairs.size();
}

void Escort::resetStats() {
	for(int i=0; i<wheelchairs.size(); ++i) wheelchairs[i].resetStats();
	collectorTime=0;
	busyTime=0;
}