/************************************************************************************
All Code by Troy Lawlor Copyright (c)2005 All Rights Reserved.
*************************************************************************************/

#include "Timing.h"

#ifdef __APPLE__
#import <Cocoa/Cocoa.h>
#endif

#ifdef WINDOWS
#include <MMSYSTEM.H>
#endif

double Time::systemTime() {
	#ifdef __APPLE__
		return CFAbsoluteTimeGetCurrent();
	#endif
	#ifdef WINDOWS
		return (double)timeGetTime()*0.001f;
	#endif
	#ifdef LINUX
		timeval time;
		gettimeofday(&time,NULL);
		double t1=time.tv_sec+(time.tv_usec/1000000.0);
		return t1;
	#else
		return 0;
	#endif
}

Timer::Timer() {
	gLastTime=gCurTime=Time::systemTime();
	gDifference=0.0f;
	timeConstant=1.0f;
}

//  call this prior to every frame drawn
void Timer::update() {
	double sysTime=Time::systemTime();
	if(sysTime-gCurTime>0.0001f) {
		gLastTime=gCurTime;
		gCurTime=sysTime;
		gDifference=gCurTime-gLastTime;
	} else {
		gDifference=0;
	}
}

double Timer::getTime() {
	return gDifference*timeConstant;
}

void Timer::setTimeConstant(float what) {
	if(what>0)
		timeConstant=what;
}

void Timer::changeTimeConstant(float amount) {
	timeConstant+=amount;
	if(timeConstant<0) timeConstant=0;
}